/*
 * ***************************************************************************
 * GAMER = < Geometry-preserving Adaptive MeshER >
 * Copyright (C) 2007-2010 -- Michael Holst and Johan Hake
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * ***************************************************************************
 */

/* ***************************************************************************
 * File:     SurfaceMesh.C    < ... >
 *
 * Author:   Johan Hake (hake.dev@gmail.com)
 *
 * Purpose:  Create and destroy SurfaceMesh data
 * ****************************************************************************
 */

#include <gamer/biom.h>
#include "gamercf.h"

/*
 * ***************************************************************************
 * Routine:  SurfaceMesh_ctor
 *
 * Author:   Johan Hake (hake.dev@gmail.com)
 *
 * Purpose:  Create a surface mesh instance
 * ***************************************************************************
 */
SurfaceMesh* SurfaceMesh_ctor(unsigned int num_vertices, unsigned int num_faces)
{
  // Allocate memory and initialize variables
  SurfaceMesh* surfmesh= (SurfaceMesh*)malloc(sizeof(SurfaceMesh));
  surfmesh->nv = num_vertices;
  surfmesh->nf = num_faces;
  surfmesh->vertex = (FLTVECT*)malloc(sizeof(FLTVECT)*surfmesh->nv);
  surfmesh->face = (INT3VECT*)malloc(sizeof(INT3VECT)*surfmesh->nf);
  surfmesh->neighbor = NULL;
  surfmesh->neighbor_list = NULL;
  surfmesh->avglen = 0.;
  surfmesh->min[0] = 0.; surfmesh->min[1] = 0.; surfmesh->min[2] = 0.;
  surfmesh->max[0] = 0.; surfmesh->max[1] = 0.; surfmesh->max[2] = 0.;
  return surfmesh;
}

/*
 * ***************************************************************************
 * Routine:  SurfaceMesh_dtor
 *
 * Author:   Johan Hake (hake.dev@gmail.com)
 *
 * Purpose:  Release memory from a SurfaceMesh instance
 * ***************************************************************************
 */
void SurfaceMesh_dtor(SurfaceMesh* surfmesh)
{
  // Free allocated memory
  free(surfmesh->vertex);
  free(surfmesh->face);

  // Destroy neighbor_list
  SurfaceMesh_destroyNeighborlist(surfmesh);
  
  free(surfmesh);
}

/*
 * ***************************************************************************
 * Routine:  SurfaceMesh_releaseData
 *
 * Author:   Johan Hake (hake.dev@gmail.com)
 *
 * Purpose:  Release data memory from a SurfaceMesh instance
 * ***************************************************************************
 */
void SurfaceMesh_releaseData(SurfaceMesh* surfmesh)
{
  // Free allocated memory
  free(surfmesh->vertex);
  free(surfmesh->face);

  // Destroy neighbor_list
  SurfaceMesh_destroyNeighborlist(surfmesh);
  
}

/*
 * ***************************************************************************
 * SubRoutine:  SurfaceMesh_createNeighborList
 *
 * Author:   Zeyun Yu (zeyun.yu@gmail.com) and Johan Hake (hake.dev@gmail.com)
 *
 * Purpose:  Create a neigbor list from a surface mesh
 * ***************************************************************************
 */              
void SurfaceMesh_createNeighborlist(SurfaceMesh* surfmesh)
{
  
  int m,n,a0,b0;
  int a,b,c;
  NPNT3 *first_ngr,*second_ngr,*tmp_ngr;
  NPNT3 **neighbor_list;

  // Destroy any exsisting neigborlist
  SurfaceMesh_destroyNeighborlist(surfmesh);

  // Create an array of NPNT3, used to store 
  neighbor_list = (NPNT3 **)malloc(sizeof(NPNT3 *)*surfmesh->nv);

  // Initialize the neighbor list
  for (n = 0; n < surfmesh->nv; n++) 
    neighbor_list[n] = NULL;

  // Iterate over the faces and collect line segments (a, b) and its connection 
  // to a face (c). Save the line segment so it forms a counter clockwise triangle 
  // with the origin vertex 
  for (n = 0; n < surfmesh->nf; n++) {
    a = surfmesh->face[n].a;
    b = surfmesh->face[n].b;
    c = surfmesh->face[n].c;
   
    first_ngr = (NPNT3 *)malloc(sizeof(NPNT3));
    first_ngr->a = b;
    first_ngr->b = c;
    first_ngr->c = n;
    first_ngr->next = neighbor_list[a];
    neighbor_list[a] = first_ngr;
    
    first_ngr = (NPNT3 *)malloc(sizeof(NPNT3));
    first_ngr->a = c;
    first_ngr->b = a;
    first_ngr->c = n;
    first_ngr->next = neighbor_list[b];
    neighbor_list[b] = first_ngr;
    
    first_ngr = (NPNT3 *)malloc(sizeof(NPNT3));
    first_ngr->a = a;
    first_ngr->b = b;
    first_ngr->c = n;
    first_ngr->next = neighbor_list[c];
    neighbor_list[c] = first_ngr;
    
  }

  // Order the neighbors so they are connected counter clockwise
  for (n = 0; n < surfmesh->nv; n++) {
    first_ngr = neighbor_list[n];
    c = first_ngr->a;
    while (first_ngr != NULL) {
      a = first_ngr->a;
      b = first_ngr->b;
      
      second_ngr = first_ngr->next;
      while (second_ngr != NULL) {
	a0 = second_ngr->a;
	b0 = second_ngr->b;
	if (a0==b && b0!=a) {
	  tmp_ngr = first_ngr;
	  while (tmp_ngr != NULL) {
	    if (tmp_ngr->next == second_ngr) {
	      tmp_ngr->next = second_ngr->next;
	      break;
	    }
	    tmp_ngr = tmp_ngr->next;
	  }
	  tmp_ngr = first_ngr->next;
 	  first_ngr->next = second_ngr;
	  second_ngr->next = tmp_ngr;
	  break;
	}
	second_ngr = second_ngr->next;
      }
      if (first_ngr->next == NULL) {
	if (first_ngr->b != c)
	{
	  printf("some polygons are not closed, Vertices: %d-%d\n ", first_ngr->b, c);
	  printf("(%f,%f,%f)\n", surfmesh->vertex[first_ngr->b].x,
		 surfmesh->vertex[first_ngr->b].y,
		 surfmesh->vertex[first_ngr->b].z);
	  printf("(%f,%f,%f)\n", surfmesh->vertex[c].x,
		 surfmesh->vertex[c].y,
		 surfmesh->vertex[c].z);
	}
      }
      
      first_ngr = first_ngr->next;
    }
  }
  
  // Attach the neigborlist to the surfmesh
  surfmesh->neighbor_list = neighbor_list;
}

/*
 * ***************************************************************************
 * SubRoutine:  SurfaceMesh_destroyNeighborlist
 *
 * Author:   Johan Hake (hake.dev@gmail.com)
 *
 * Purpose:  Release memory from a neigborlist
 * ***************************************************************************
 */
void SurfaceMesh_destroyNeighborlist(SurfaceMesh* surfmesh)
{
  unsigned int n;
  NPNT3* first_ngr = NULL;
  NPNT3* tmp_ngr = NULL;

  if (surfmesh->neighbor_list != NULL)
  {
    // Release the single neighbors 
    for (n = 0; n < surfmesh->nv; n++) {
      first_ngr = surfmesh->neighbor_list[n];
      while (first_ngr != NULL) 
      {
	tmp_ngr = first_ngr->next;
	free(first_ngr);
	first_ngr = tmp_ngr;
      }
    } 

    // Free the array of pointers
    free(surfmesh->neighbor_list);
    surfmesh->neighbor_list = NULL;
  }
}

